;FILE: FSK1200A.ASM, Last edition: 31-APR-1997
;1200 bps FSK modem for standard terrestial VHF packet.
;(c) 1995-1997 Pawel Jalocha, SP9VRC,
;e-mail: jalocha@chopin.ifj.edu.pl, Pawel.Jalocha@cern.ch
;
;The FSK demodulator here is non-synchronous, based on two filters
;matched to the FSK tones. The mark/space decision is done by comparing
;the tone's amplitudes, thus the modem performs best when both tones
;have equal amplitudes.
;The FSK modulator is traditional continous-phase FSK keying
;without any shaping.
;
;This software is not to be used for purposes other than amateur radio
;without a written permision of the author.
;Usage for profit or in commercial/military products is explicitly prohibited.

        page 132
        nolist
	include 'leonid'
	list

        scsjmp short    ;force short jumps for .if/.endi, .while/.endw, etc.
;        opt nops        ;do not pack strings (make one character per word)

        title '1200 bps FSK modem by SP9VRC'

;*************************************************************************
;compile-time constants

;the EVM56K is now defined in leonid.asm
;EVM56K  equ 0           ;0 => DSPCARD4
                        ;1 => EVM56002

SampleFreq equ  9600.0 ;[Hz]    ;9600 Hz sampling

AliasFilterLen equ 32 ;[samples];the length of the anti-alias FIR filter
AliasLowFreq    equ  600.0 ;[Hz];lower and upper edges of the input/output filters
AliasUppFreq    equ 2800.0 ;[Hz]
DecimateRatio   equ 1           ;decimation ratio


SymbolRate      equ 1200 ;[Hz]  ;bits/second
SymbolLen       equ @cvi(SampleFreq/SymbolRate/DecimateRatio+0.5) ;[samples]
                                ;the time-spacing of our symbols
SymbolLenLog2   equ @cvi(@log(SymbolLen)/@log(2)+0.5)

LowToneFreq       equ 1200 ;[Hz] ;Lower FSK tone frequency
HighToneFreq      equ 2200 ;[Hz] ;Higher FSK tone frequency

TxToneAmpl      equ  0.25       ;amplitude of the Tx tone which is sent out
                                ;this should be as high as possible but
                                ;if too high the output of the anti-alias
                                ;filter will go out of CODEC's range.
TxAttenuate     equ  10.0 ;[dB] attenuate the output CODEC's signal.
                                ;Set this between 0.0 and 90.0
                                ;to accomodate the output level for the
                                ;modulator input of your tranceiver.
                                ;Experiments indicate that when you attenuate
                                ;here the signal is noisy and distorted
                                ;thus it is recomended to use an external
                                ;attenuator.

TxDelay       equ 20    ;default value for the TxDelay (20 => 200 ms)
TxTail        equ 3     ;default value for the TxTail (3 => 30 ms)
TxDuplex      equ 0     ;1 = full duplex mode (for satellite uplink)
                        ;0 = simplex mode (for HF or VHF packet)
SciTimerRate equ 19200  ;interrupt rate of the SCI timer. This we need
                        ;for proper scaling of TxDelay and TxTail.
                        ;It is 19200 unless you have messed up the LEONID
                        ;or the SCI itself or if the new LEONID version
                        ;has changed the data/timer rate.

RxGain          equ 10.0 ;[dB] Rx CODEC's input gain
                                ;set this between 0 and 22.5 dB depending on
                                ;the audio level from your tranceiver
RxAGC           equ 1           ;enable the automatic gain adjustment
                                ;for the CODEC's audio input.
                                ;ideally you should avoid the AGC and set
                                ;RxGain for best audio level match.
                                ;Note, that the AGC may get fooled up by
                                ;for example long periods of total silence,
                                ;thus you should rather keep the quelch (if any)
                                ;open all the time.
RxAGC_MSfollow  equ 1.0/128.0   ;update weight per dec. sample
                                ;for the Mean Square tracking
RxAGChold       equ 2400 ;[Dec. samples]  AGC peak hold time
RxAGCfall       equ 240  ;[Dec. samples]  AGC falling period/1.5 dB step
RxAGCminMS      equ 0.008       ;min. and max. tolerable audio Mean Square level
RxAGCmaxMS      equ 0.04
RxAGCavoidDCD   equ 0           ;Disallow gain changes when DCD is on
                                ;so the gain changes do not mess up
                                ;the transmition being received.
RxAGCavoidPTT   equ 0           ;The AGC would ignore the silence on the Rx
                                ;when the PTT is on. Usefull for simplex
                                ;but not good for full duplex.

RxPowerDiffDist equ 2           ;distance in samples to differentiate
                                ;the tone's power for symbol sync.
RxSymbolSyncWeight equ 5 ;[shifts] ;Weight factor for tracking
                                ;of the symbol sync. and DCD
                                ;higher number => slower but a more stable
                                ;tracking

RightChannel    equ 0           ;Use the right CODEC's channel not the left.
                                ;I didn't try this but Timo says it works.

NewBaudRate     equ 0           ;Set a new baud rate on the serial port
                                ;for faster KISS or SPYing
                                ;When 0 the serial port runs at BIOS'es rate.

BufLen   equ      64    ;Sample buffer length with some safety margin
                        ;but not too much because large BufLen make extra
                        ;delays in the transmitions (we have to hold the PTT
                        ;until this buffer is flushed) so one can get some
                        ;improvement in the turn-around time by optimizing
                        ;BufLen.
                        ;BufLen must not be smaller than
                        ;AliasFilterLen+DecimateRatio
                        ;Also: smaller BufLen makes the code intolerant
                        ;to sudden processing demand (like to process
                        ;a bigger FEC block or to send out a long message).

SPY      equ 0  ; SPY=0 => KISS mode
                ; SPY=1 => KISS code excluded, selected debug code included
SPY_BlockLen equ 512    ;how many words to follow each SPY request,
                        ;originally 512 but can be changed here.
                        ;If you are spying on very fast data don't put more
                        ;than what the serial output buffer can hold.

SPY_RxIQ        equ 0   ;I/Q after the Hilbert transformer
SPY_AudioMS     equ 0   ;Signal's Mean Square seen by the AGC
SPY_CodecGain   equ 0   ;CODEC's gain being changed by the AGC
SPY_RxCodecPtr  equ 0   ;CODEC's pointer
SPY_RxLowToneIQ equ 0   ;I/Q after mixing with the low tone
SPY_RxHighToneIQ equ 0  ;I/Q after mixing with the high tone
SPY_RxLowToneFiltIQ equ 0 ;as above but after the symbol intergrator
SPY_RxHighToneFiltIQ equ 0
SPY_RxTonePower equ 0   ;the power of the lower and higher tones
SPY_RxTonePowerLog2 equ 0   ;the Log2 of the tone powers
SPY_RxTonePowerDiff equ 0   ;differenciated tone's powers
SPY_RxSymbolSync    equ 0   ;symbol sync. tap
SPY_RxSymbolSamplePtr equ 0
SPY_RxDCDlevels     equ 1

MonitorRxAGC    equ 0           ;monitor the gain changes being done by the AGC
                                ;with the UP and DOWN LEDs

;*************************************************************************
;Some simple macros

      if RightChannel
PTT     macro mode      ;PTT line: clr/set/chg
        b\mode #4,X:$FFE4
        endm
      else
PTT     macro mode      ;PTT line: clr/set/chg
        b\mode #0,X:$FFE4
;      b\mode #3,X:$FFE4       ;*** DEBUG ***
        endm
      endif

      if RightChannel
PushUp  macro mode      ;Push UP line of the TRX
        b\mode #5,X:$FFE4
        endm
      else
PushUp  macro mode
        b\mode #1,X:$FFE4
        endm
      endif

      if RightChannel
PushDown macro mode      ;Push DOWN line of the TRX
        b\mode #6,X:$FFE4
        endm
      else
PushDown macro mode
        b\mode #2,X:$FFE4
        endm
      endif

UpLED   macro mode      ;UP line (a red LED connected)
        b\mode #1,X:$FFE4
        endm

DownLED macro mode      ;DOWN line (a red LED connected)
        b\mode #2,X:$FFE4
        endm

YellowLED macro mode    ;CAT line (a yellow LED connected)
        b\mode #3,X:$FFE4
        endm

RedLED  macro mode      ; Red LED clr/set/chg
        b\mode #13,X:$FFE4
        endm

LongDecayAver macro Input,Aver,Scale
              sub Aver,Input
              rep Scale
                asr Input
              add Input,Aver
              endm

;*************************************************************************
;The actuall code (internal/external program RAM)

        LOMEM P:$0000
        HIMEM P:$1FFF

        org     p:user_code

        jmp <Initialize

ProcessLoop     ;main loop to process CODEC's samples

        move #>DecimateRatio,a    ;a = how many samples we should have before
        move X:<RxCodecPtr,r2     ;proceeding
        jsr <WaitSampleBlock      ;wait for that many samples to come
        move #BufLen*4-1,m2       ;set m2 so r2 wraps correctly
        move #<4,n2
        if SPY&&SPY_RxCodecPtr
          move r2,a
          rep #8
            asl a
          jsr <SpyA
        endif

;the receiver code
        move #<AliasFilterLen-1,m4
      if RightChannel
        move (r2)+
      endif                        ;filter, decimate and fill the window's tap
          move #4*(AliasFilterLen-DecimateRatio),n2
          move #AliasFilterInpI,r4 ;r4 to address the input filter shape
          move (r2)-n2            ;move back in the CODEC input buffer
          move #<4,n2             ;n2=4 for easier input adressing
          move r2,x1              ;save r2 (where we start the filter tap)
          clr a X:(r2)+n2,x0 Y:(r4)+,y0
          .loop #AliasFilterLen-1
            mac x0,y0,a X:(r2)+n2,x0  Y:(r4)+,y0
          .endl
          macr x0,y0,a #AliasFilterInpQ,r4
          move x1,r2      ;restore the tap start
          nop
          clr b X:(r2)+n2,x0 Y:(r4)+,y0
          .loop #AliasFilterLen-1
            mac x0,y0,b X:(r2)+n2,x0  Y:(r4)+,y0
          .endl
          macr x0,y0,b a,X:<RxTmpI ;a,b = decimated sample (I/Q)
          move b,X:<RxTmpQ
        if SPY&&SPY_RxIQ
          jsr <SpySync
          .if <cc>
            jsr <SpyA
            tfr b,a a,L:<SPY_Scratch
            jsr <SpyA
            move L:<SPY_Scratch,a
          .endi
        endif
      if RightChannel
        move (r2)-
      endif
;        move r2,X:<RxCodecPtr     ;save the CODEC buffer pointer
                                ;no, we do it after the output interpolation

      if RxAGC                  ;update the signals power (mean square)
       if RxAGCavoidPTT
        PTT tst         ;avoid any AGC level measurement or gain change
        jcs <CheckMS_OK ;while PTT is on.
       endif
        move X:<RxTmpI,x0       ;compute the MS of the input audio
        mpy x0,x0,a X:<RxTmpQ,x0
        macr x0,x0,a X:<RxAudioMS,x1
        move a,x0 #RxAGC_MSfollow,y0
        mpy x0,y0,a #1.0-RxAGC_MSfollow,y0
        macr x1,y0,a
        move a,X:<RxAudioMS

       if SPY&&SPY_AudioMS
        jsr <SpyA
       endif

      if MonitorRxAGC
        UpLED clr
        DownLED clr
      endif

      if RxAGCavoidDCD
       jset #2,X:<RxState,CheckMS_OK
      endif

      move X:<RxCodecPtr,r2     ;set r2 to address the CODEC's buffer
      move #BufLen*4-1,m2       ;set m2 so r2 wraps in the CODEC's buffer
                                ;we will need r2/m2 for RxGainUp/Down

        move #RxAGCminMS,x0     ;see, if the MS is below the minimal level
        cmp x0,a
        jcc <CheckMSmax         ;if not, then check if above the max. level
                                ;if so, increase the CODEC's input gain
        move X:<RxAGCcount,a    ;decrement the timeout
        move #>1,x0
        sub x0,a 
        move a,X:<RxAGCcount
        jgt <CheckMS_OK         ;leave if not yet zero
        jsr <RxGainUp           ;increase the CODEC's gain
        jcs <CheckMS_OK
        move #0.7071,y0         ;increase the MS to follow
        move X:<RxAudioMS,x0    ;the gain change faster
        mpyr x0,y0,a #>RxAGCfall,x0 ;preset the AGC "timer" to prohibit
        asl a x0,X:<RxAGCcount      ;gain increase for the RxAGCfall period
        move a,X:<RxAudioMS
      if MonitorRxAGC
        UpLED set
      endif
        jmp <CheckMS_OK

CheckMSmax                      ;is the MS above the given maximum
        move #>RxAGChold,y0     ;preset the AGC "timer" to prohibit gain increase
        move #RxAGCmaxMS,x0     ;for the RxAGChold period
        cmp x0,a y0,X:<RxAGCcount ;MS above maximum ?
        jcs <CheckMS_OK         ;if not then we are done
GainDown
        jsr <RxGainDown         ;otherwise decrease the CODEC's gain
        jcs <CheckMS_OK
        move #0.7071,y0         ;decrease the MS to follow expected
        move X:<RxAudioMS,x0    ;gain reduction faster
        mpyr x0,y0,a #>RxAGChold,x0     ;initialize the AGC hold count-down
        move a,X:<RxAudioMS
        move x0,X:<RxAGCcount
      if MonitorRxAGC
        DownLED set
      endif

CheckMS_OK

      endif     ;of RxAGC

        ;here the Rx should process the I/Q sample in X:<RxTmpI/Q

        move X:<RxHighTonePhase,x0  ;get the high tone's phase
        move X:<RxHighToneFreq,a    ;advance the phase
        move            #<SymbolLen-1,m1
        add x0,a        X:<RxToneFiltPtr,r1
        move a1,X:<RxHighTonePhase
        jsr <IQ               ;compute I and Q (modifies a,b,x,y,r0,m0,n0)
        move a,y1             ;mix the Rx I/Q sample with the tone's carrier
        move b,y0
        move X:<RxTmpI,x1
        mpy x1,y1,a X:<RxTmpQ,x0
        macr x0,y0,a    #<RxHighToneFilt-RxLowToneFilt,n1
        mpy x1,y0,b     
        macr -x0,y1,b a,X:(r1+n1) ;a/b = high tone mixer output (I/Q)
        move b,Y:(r1+n1)         ;put a/b into the filter's tap
      if SPY&&SPY_RxHighToneIQ
        jsr <SpySync
        .if <cc>
          jsr <SpyA
          tfr b,a a,L:<SPY_Scratch
          jsr <SpyA
          move L:<SPY_Scratch,a
        .endi
      endif

        move X:<RxLowTonePhase,x0  ;get the low tone's phase
        move X:<RxLowToneFreq,a    ;advance the phase
        add x0,a
        move a1,X:<RxLowTonePhase
        jsr <IQ               ;compute I and Q (modifies a,b,x,y,r0,m0,n0)
        move a,y1             ;mix the Rx I/Q sample with the tone's carrier
        move b,y0
        move X:<RxTmpI,x1
        mpy x1,y1,a X:<RxTmpQ,x0
        macr x0,y0,a
        mpy x1,y0,b
        macr -x0,y1,b a,X:(r1)  ;a/b = low tone mixer output (I/Q)
        move b,Y:(r1)+          ;a/b to the filter's tap and advance the ptr
        move r1,X:<RxToneFiltPtr
      if SPY&&SPY_RxLowToneIQ
        jsr <SpySync
        .if <cc>
          jsr <SpyA
          tfr b,a a,L:<SPY_Scratch
          jsr <SpyA
          move L:<SPY_Scratch,a
        .endi
      endif

      move L:(r1)+,ab           ;sum up the low tone filter's tap
      move L:(r1)+,x            ;we are doing a boxcar filter
      .loop #SymbolLen-2        ;but we could put here any shape
        add x1,a X:(r1),x1
        add x0,b Y:(r1)+,x0
      .endl
      add x1,a
      add x0,b (r1)+n1          ;a/b = low tone filter output
      rep #SymbolLenLog2-1
        asr a
      rep #SymbolLenLog2-1
        asr b

        move a,y0               ;save the low tone filter output in y0/1
        move b,y1

      if SPY&&SPY_RxLowToneFiltIQ
        jsr <SpySync
        .if <cc>
          jsr <SpyA
          tfr b,a a,L:<SPY_Scratch
          jsr <SpyA
          move L:<SPY_Scratch,a
        .endi
      endif

      move L:(r1)+,ab           ;sum up the high tone filter's tap
      move L:(r1)+,x
      .loop #SymbolLen-2
        add x1,a X:(r1),x1
        add x0,b Y:(r1)+,x0
      .endl
      add x1,a
      add x0,b (r1)-n1          ;a/b = high tone filter output
      rep #SymbolLenLog2-1
        asr a
      rep #SymbolLenLog2-1
        asr b

      if SPY&&SPY_RxHighToneFiltIQ
        jsr <SpySync
        .if <cc>
          jsr <SpyA
          tfr b,a a,L:<SPY_Scratch
          jsr <SpyA
          move L:<SPY_Scratch,a
        .endi
      endif

        mpy y0,y0,a a,x1
        move y1,y0
        mac y0,y0,a b,x0        ;a = low tone's power
        mpy x0,x0,b x1,x0
        mac x0,x0,b             ;b = high tone's power

        move X:<RxTonePowerPtr,r1
        move #<2*SymbolLen-1,m1
        move #<RxPowerDiffDist,n1
        move ab,L:-(r1)
        move L:(r1+n1),x
        sub x0,b                ;b = high tone power difference
        sub x1,a                ;a = low tone power difference
        move r1,X:<RxTonePowerPtr

      if SPY&&SPY_RxTonePowerDiff
        jsr <SpySync
        .if <cc>
          jsr <SpyA
          tfr b,a a,L:<SPY_Scratch
          jsr <SpyA
          move L:<SPY_Scratch,a
        .endi
      endif

      if SPY&&SPY_RxTonePower
        jsr <SpySync
        .if <cc>
          move a,L:<SPY_Scratch
          move X:(r1),a
          jsr <SpyA
          move Y:(r1),a
          jsr <SpyA
          move L:<SPY_Scratch,a
        .endi
      endif

      if SPY&&SPY_RxTonePowerLog2       ;this SPY changes a,b,x,y,r0
        jsr <SpySync
        .if <cc>
          move X:(r1),a
          jsr <Log2
          rep #6
            asr a
          jsr <SpyA
          move Y:(r1),a
          jsr <Log2
          rep #6
            asr a
          jsr <SpyA
        .endi
      endif

        move X:<RxSymbolSyncPtr,r4      ;update the symbol sync. data
        move #<SymbolLen-1,m4
        move a,x0                       ;take the low tone power diff.
        mpy x0,x0,a b,x0                ;square it
        mac x0,x0,a L:-(r4),b           ;add the high tone power diff.
        LongDecayAver a,b,#RxSymbolSyncWeight
        move b,L:(r4)
        move r4,X:<RxSymbolSyncPtr
      if SPY&&SPY_RxSymbolSync
        tfr b,a
        jsr SpyA
      endif

        move r4,a               ;if the symbol sync. pointer is same
        move X:<RxSymbolSamplePtr,x0  ;as the sampling pointer
        eor x0,a1 #<SymbolLen,n1
        .if <eq>                ;then sample the data bit
          move (r1)+
          move L:(r1),x
          move L:(r1+n1),y
          tfr x1,a (r1)-
          sub x0,a
          tfr y1,b
          sub y0,b a1,x0
          eor x0,b1
          not b1
          btst #23,b1
          if !SPY
            putbit
            nop
          endif
        .endi

        move #<SymbolLen/2,n4
        move X:<RxSymbolSamplePtr,x0
      if SPY&&SPY_RxSymbolSamplePtr
        tfr x0,a #>RxSymbolSyncTap,y0
        sub y0,a
        rep #10
          asl a
        jsr <SpyA
      endif
        lua (r4)+n4,r5
        move m4,m5
        move r5,a        ;if sampling pointer is oposite the sync. pointer
        eor x0,a1        ;then adjust the sampling point
        .if <eq>
          move L:(r4)+,a
          move L:(r5)+,y
          sub y,a L:(r4)-,b
          move L:(r5)-,y
          sub y,b
          cmp a,b
          .if <gt>
            move (r5)+
            move r5,X:<RxSymbolSamplePtr
          .else
            move L:-(r4),b
            move L:-(r5),y
            sub y,b (r4)+
            cmp a,b (r5)+
            .if <gt>
              move (r5)-
              move r5,X:<RxSymbolSamplePtr
            .endi
          .endi

          move r5,r4            ;DCD check
        if SPY&&SPY_RxDCDlevels
          jsr <SpySync
          .if <cc>
            move L:(r5),a
            rep #4
              asl a
            jsr <SpyA
            move L:(r4+n4),a
            rep #4
              asl a
            jsr <SpyA
          .endi
        endif
          move L:(r5),a
          asl a L:(r4+n4),b
          btst #2,X:<RxState    ;DCD is on ?
          .if <cs>              ;when it is on
            cmp a,b             ;if two little signal
            .if <le>            ;we declare the DCD is off
               if SPY
                 caroff
               endif
               bclr #2,X:<RxState
               RedLED clr
            .endi
          .else                 ;when DCD is off
            move L:(r5),x
            add x,a
            cmp a,b
            .if <gt>
               if SPY
                 caron
               endif
               bset #2,X:<RxState
               RedLED set
            .endi
          .endi

        .endi

;the transmitter code

        ;here the Tx should make ready the new sample in a,b

        move #>1,x0
        move X:<TxSymbolTime,a
        sub x0,a #>SymbolLen,x0
        move a,X:<TxSymbolTime
        .if <le>                        ;when the time has come for a new bit
          move x0,X:<TxSymbolTime
          if SPY                        ;when if SPY mode
            move #>1,x0
            move X:<TxTestCounter,a
            sub x0,a
            move a,X:<TxTestCounter
            .if <le>
              PTT chg
              move #>1200,a
              .if <cc>
                asl a #>$7F7F7F,x0
                asl a x0,X:<TxTestReg
              .endi
              move a,X:<TxTestCounter
            .endi
            move X:<TxTestReg,a1
            lsr a
            .if <eq>
              jsr <Rand48
              move a0,x0
              eor x0,a1
              ori #$01,ccr
              ror a1
            .endi
            .if <cc>
              bchg #1,X:<TxState
            .endi
            move a1,X:<TxTestReg
          else                          ;if in true (non-SPY) mode
            getbit                      ;get nest bit to be transmitted
            .if <eq>                    ;if <eq> then no more bits
              bchg #1,X:<TxState        ;then flip the level
            .else                       ;but if the bits are there then
             .if <cc>                   ;if bit==0 then
               bchg #1,X:<TxState       ;flip the level
             .endi                      ;otherwise don't change it
            .endi
          endif
        .endi

        clr a #<0,b
        PTT tst                 ;if PTT is off
        jcc <TxSample           ;then we transmit silence
        move #>2.0*DecimateRatio*LowToneFreq/SampleFreq,a
        btst #1,X:<TxState      ;which tone we should transmit ?
        .if <cs>                ;change if the high tone
          move #>2.0*DecimateRatio*HighToneFreq/SampleFreq,a
        .endi
        move X:<TxTonePhase,x0  ;get the Tx tone's phase
        add x0,a                ;advance the phase by the frequency
        move a1,X:<TxTonePhase
        jsr <IQ                 ;compute I and Q (modifies a,b,x,y,r0,m0,n0)
        move a,x1 #>TxToneAmpl,y0   ;multiply by the TxToneAmpl
        mpyr x1,y0,a b,x0
        mpyr x0,y0,b
TxSample                        ;a,b = Tx sample (I/Q) ready to be output
        move X:<RxCodecPtr,r2
        move #BufLen*4-1,m2
        move #AliasFilterOutI,r4
        move #<AliasFilterLen-1,m4
        move #<AliasFilterOutQ-AliasFilterOutI,n4
        move (r2)+
      if RightChannel
        move (r2)+
      endif                     ;output interpolation
          move #4*(AliasFilterLen-DecimateRatio),n2
          move a,y1
          move b,y0
          move (r2)-n2            ;move back in the CODEC output buffer
          move #<4,n2             ;n2=4 for easier output adressing
          .loop #AliasFilterLen-DecimateRatio
            move X:(r4),x0 Y:(r2),a
            mac x0,y1,a X:(r4+n4),x0
            macr x0,y0,a (r4)+
            move a,Y:(r2)+n2      ;output to the left channel
          .endl
          .loop #DecimateRatio
            move X:(r4),x0
            mpy x0,y1,a X:(r4+n4),x0
            macr x0,y0,a (r4)+
            move a,Y:(r2)+n2
          .endl
          nop
        move (r2)-
      if RightChannel
        move (r2)-
      endif
        move r2,X:<RxCodecPtr

        jmp     <ProcessLoop

;------------------------------------------------------------------------------
;CODEC's buffer synchronization to the sample stream

CheckSampleBlock        ;a = how many samples we want
        asl a           ;r2 = where we wait for the block to start
        asl a #>2,x1
        add x1,a
        move a,x1
CheckSampleBlock_1
        move r7,a
        move r2,x0
        sub x0,a
        jpl <CheckSample_cmp
          move #>BufLen*4,x0
          add x0,a
CheckSample_cmp
        cmp x1,a
        rts             ;on output: a,x0,x1 are modified
                        ;If carry=0 => we are sure that:
                        ;1. there are at least A new input samples at X:(r2)
                        ;   to be read
                        ;2. there are at least A locations at Y:(r2) to
                        ;   be written


WaitSampleBlock         ;wait for a block of samples from the CODEC
                        ;on input: a = how many samples we want
                        ;          r2 = where the block starts
        jsr <CheckSampleBlock
        jcc <WaitSample_ret
WaitSample_loop
        wait            ;wait (at low power) until an interrupt comes
        jsr <CheckSampleBlock_1
        jcs <WaitSample_loop
WaitSample_ret
        rts             ;on output: a,x0,x1 are modified
                        ;We are sure that:
                        ;1. there are at least A new input samples at X:(r2)
                        ;   to be read
                        ;2. there are at least A locations at Y:(r2) to
                        ;   be written

;------------------------------------------------------------------------------
;Trigonometric functions

PI      equ     3.14159265358979323846

;this routine computes a cosine/sine pair using the sine ROM
;with a second order (linear+quadrature) approximation between table points
IQ                              ;x0 = angle ( -1 = -PI, +1 = +PI)
;        ori #%00000100,omr      ;enable the sine ROM table
        move #>$80,x1   ;shift out 8 most significant bits
        mpy x0,x1,a  #>$FF,x0
        move x0,m0
        and x0,a     #>$100,x0
        or x0,a      #<$40,n0
        move a1,r0      ;put the 8 most significant bits into r0 with offset = $100
        move a0,y0      ;save the remaining bits in y0
        jclr #23,y0,SinTable_lev2
          move (r0)+
SinTable_lev2
        move Y:(r0+n0),x0       ;x0 = coarse cosine
        move Y:(r0),x1          ;x1 = coarse sine
        mpyr x1,y0,a  #PI/256.0,y1
        tfr x0,a  a,x1
        macr -x1,y1,a           ;a = fine cosine
        mpyr x0,y0,b  Y:(r0),x1
;        andi #%11111011,omr     ;disable the sine ROM table
        tfr x1,b  b,x1
        macr x1,y1,b  #PI*PI/2.0/65536.0,y1  ;b = fine sine
        mpyr y0,y0,a  a,x0
        move a,y0
        mpyr y0,y1,a
        tfr x0,a  a,y1
        macr -x0,y1,a  b,x1     ;a = super fine cosine
        macr -x1,y1,b           ;b = super fine sine
        rts                     ;x,y are modified
                                ;r0,m0,n0 are modified
                                ;maximum error is about 0.7E-6
                                ;execution time 4+64+4 clock cycles
                                ;including "jsr <IQ_lev2" and "rts"

;------------------------------------------------------------------------------
;Random number generator
      if SPY
Rand48                          ;a simple random number generator
        move L:<Rand,a
        asl a L:<RandInc,x
        adc x,a
        move a10,L:<Rand
        rts                     ;returns a pseudo-random number in a
                                ;modifies x
      endif
;------------------------------------------------------------------------------
;interface to the LEONID's KISS paramaters and the PTT logic

        if !SPY

; KISS control frame handling - called by LEONID when a KISS control-type
; frame is received with non-standard parameters
KISSctrl        ;a = parameter number, x1 = parameter value
                ;we should only use a,x0,x1,r3
        rts

; transmitter PTT control - a routine called by LEONID to say that we
; should transmit (carry=1) or stop transmitting (carry=0)
PTTctrl jcc <PTT_off
        PTT set
        rts
PTT_off PTT clr
        rts

        endif
;------------------------------------------------------------------------------

        if RxAGC

RxGainUp                        ;increase the CODEC's input gain (both channels)
        clr a #>$0F0F00,x0      ;r2/m2 should point to the CODEC's buffer
        move Y:(r2),a1          ;get the CODEC's input control word
        and x0,a                ;extract the gain bits
	cmp x0,a  #>$010100,x0  ;already maximum ?
        jeq <RxGainChange_abort ;if so then abort this attempt
	add x0,a  #>$F0F000,x0  ;if not, increment the gain by 1
	move a1,x1
	move Y:(r2),a1          ;and reload all the control words
	and x0,a  n2,x0         ;in the output buffer
	or x1,a  #<4,n2         ;make n2=4 for a moment
	.loop #BufLen
	  move a1,Y:(r2)+n2
	.endl
	move x0,n2              ;restore n2
        andi #$FE,ccr           ;clear carry
        rts                     ;modifies a,x0,x1

RxGainDown                      ;decrease the CODEC's input gain (both channels)
        clr a #>$0F0F00,x0      ;r2/m2 should point to the CODEC's buffer
        move Y:(r2),a1          ;get the CODEC's input control word
        and x0,a  #>$010100,x0  ;extract the gain bits
        sub x0,a  #>$F0F000,x0  ;attempt to decrease the gain
        jcs <RxGainChange_abort ;jump if overflow
        move a1,x1
        move Y:(r2),a1          ;reload all the input control words
        and x0,a  n2,x0         ;in the buffer with the new input gain
        or x1,a  #<4,n2         ;n2=4 for a moment
        .loop #BufLen
          move a1,Y:(r2)+n2
        .endl
        move x0,n2              ;restore n2
        andi #$FE,ccr           ;clear carry
        rts                     ;modifies a,x0,x1

RxGainChange_abort
        ori #$01,ccr            ;set carry to indicate that we couldn't
        rts                     ;change the gain

        endif

;------------------------------------------------------------------------------

      if SPY
Log2                    ;a2:1:0 = 56-bit argument (0.0,+256.0)
	tst a #<0,r0    ;r0 ready for finding the integer part
	jmi <Log2_Neg   ;if argument negative...
	jeq <Log2_Inf   ;if zero...
	jnr <Log2_Frac  ;if already normalized
Log2_Norm               ;normalize the argument to the range <0.5,1.0)
	rep #8
	  norm r0,a     ;r0 counts the number of shifts performed
	jnn <Log2_Norm
			;r0 = approx. integer result
Log2_Frac               ;a = argument scaled to range <0.5,1.0) = X
	clr a a,x0      ;x0 = X
	move r0,a2      ;now we compute the finer fractional approximation
	asr a x0,x1     ;a = approx. integer part
	bchg #23,x0     ;x0 = X-1
	add x0,a        ;a += 2*(X-1) => first, linear correction
	add x0,a
	bchg #22,x1     ;x1 = X-1/2
	mpyr x0,x1,b #0.371,x1 ;b = (X-1)*(X-1/2)
	sub b,a b,x0    ;a += (X-1)*(X-1/2)
	macr -x0,x1,a   ;a += 0.371*(X-1)*(X-1/2)
	rts             ;a = log2(input) <-48.0,+8.0) => second, parabolic corr.
			;b,x,r0 are modified
			;error RMS is 0.005, maximum error is 0.008
Log2_Neg                ;for zero and negative arguments we give the most
Log2_Inf clr a          ;negative result: a = -48.0
	move #<$E8,a2
	rts             ;execution time is about 80 clock cycles
			;but becomes larger for very small arguments
      endif
;------------------------------------------------------------------------------

Initialize      ;initialize registers, buffers, windows, etc.

        ori #%00000100,omr      ;enable the sine ROM table
                                ;this is for the IQ routine

      if !SPY
        move #>@cvi(TxDelay*SciTimerRate/100.0),a ;set TxDelay
        move a,p:kiss_pars
        move #>@cvi(TxTail*SciTimerRate/100.0),a ;set TxTail
        move a,p:kiss_pars+3
        move #<TxDuplex,a1                      ;set TxDuplex
        move a1,p:kiss_pars+4
      endif

        move #$FFFF,m1
        clr a #<RxLowToneFilt,r1
        rep #SymbolLen
          move a,L:(r1)+
        clr a #<RxHighToneFilt,r1
        rep #SymbolLen
          move a,L:(r1)+
        move #<RxTonePowerPipe,r1
        rep #2*SymbolLen
          move a,L:(r1)+
        move #<RxSymbolSyncTap,r1
        rep #SymbolLen
          move a,L:(r1)+

      if !SPY
        move #>KISSctrl,a        ;switch serial interface to KISS mode
        move #>PTTctrl,b
        opensc
      endif
      if NewBaudRate        ;set new baud rate (copied from LEONID)
        movep #(xtal+2*16*NewBaudRate)/(2*2*16*NewBaudRate)-1,X:<<$FFF2
      endif

        move #Buffer+2,r7        ;for the CODEC's interrupt routine
        move #BufLen*4-1,m7

        move #Buffer,r2          ;for "cdctrl" to initialize the buffer
        move #<4-1,n2
        move #BufLen*4-1,m2
                        ;initialize input/output control words in the buffer
                        ;zero input/output data
      if EVM56K         ;for EVM56002 use MIC input
        ctrlcd  1,r2,BufLen,MIC,RxGain,RxGain,LINEO|HEADP,TxAttenuate,TxAttenuate
      else              ;for DSPCARD4 use LINE input
        ctrlcd  1,r2,BufLen,LINEI,RxGain,RxGain,LINEO|HEADP,TxAttenuate,TxAttenuate
      endif
        opencd SampleFreq/1000.0,HPF     ;start taking samples at given rate

        jmp <ProcessLoop

        if SPY

SpySync move a10,L:<SpySave     ;output: carry=1 => no spy request
        move a2,X:<SpySave+1    ;carry=0 => spy request !
        move x0,Y:<SpySave+1    ;SPY_BlockLen words (jsr <SpyA) should follow
        move x1,Y:<SpyCount
        move X:<SpyCount,a
        tst a
        jne <SpyCont
        lookc 0
        jcs <Spy_end
        move #>'S',a
        cmp x0,a
        ori #$01,ccr
        jne <Spy_end
        move #>'P',x0
        putc
        move #>SPY_BlockLen,a
        move a,X:<SpyCount
SpyCont andi #$FE,ccr
        jmp <Spy_end

SpyA    move a10,L:<SpySave
        move a2,X:<SpySave+1
        move x0,Y:<SpySave+1
        move x1,Y:<SpyCount
        move X:<SpyCount,a
        tst a
        jne <Spy_copy

Spy_check
        lookc 0
        jcs <Spy_end
        move #>'S',a
        cmp x0,a
        jne <Spy_end
        move #>'P',x0
        putc
        move #>SPY_BlockLen,a
Spy_copy
        move #>1,x0
        sub x0,a
        move a,X:<SpyCount

        move X:<SpySave,a
        rep #8
          lsr a
        move a1,x0
	putc
        move X:<SpySave,a
        rep #16
          lsr a
        move a1,x0
        putc

Spy_end move L:<SpySave,a10
        move X:<SpySave+1,a2
        move Y:<SpySave+1,x0
        move Y:<SpyCount,x1
        rts

        endif

;*************************************************************************
;Internal data RAM

        LOMEM X:$0000,Y:$0000,L:$0000
        HIMEM X:$00FF,Y:$00FF,L:$00FF

        org L:user_data

        if SPY
SpySave dc 0,0
SpyCount dc 0
SPY_Scratch dc 0
        endif

RxTmpI  ds 1    ;temporary storages for the receiver's symbol decoder
RxTmpQ  ds 1

      if SPY
Rand    dc $954820584783        ;seed and increment for the random
RandInc dc $894659236475        ;number generator
      endif

LastL = *
        org X:LastL
        org Y:LastL

        org X:                  ;internal X-RAM

RxCodecPtr    dc Buffer

        if RxAGC
RxAudioMS       dc 0            ;Mean Square (the power) of the input audio
RxAGCcount      dc RxAGCfall
        endif

RxState         dc 0            ;bit #0 = ?
                                ;bit #1 = ?
                                ;bit #2 = DCD decision
                                ;bit #3 = momentary DCD decision

RxLowToneFreq   dc 2.0*DecimateRatio*LowToneFreq/SampleFreq
RxLowTonePhase  dc 0
RxHighToneFreq  dc 2.0*DecimateRatio*HighToneFreq/SampleFreq
RxHighTonePhase dc 0

RxToneFiltPtr   dc RxLowToneFilt
RxTonePowerPtr  dc RxTonePowerPipe
RxSymbolSyncPtr dc RxSymbolSyncTap
RxSymbolSamplePtr dc RxSymbolSyncTap

TxState         dc 0    ;bit #0 = 1 => transmit, 0 => Tx quiet
                        ;bit #1 = actual tone (high/low)
TxSymbolTime    dc SymbolLen
TxTonePhase     dc 0
        if SPY
TxTestCounter   dc 1000
TxTestReg       dc 0
        endif

        org Y:          ;internal Y-RAM


        org X:
LastX = *
        org Y:
LastY = *
        if @cvi(LastX)>=@cvi(LastY)
          org L:LastX
        else
          org L:LastY
        endif

        org L:

RxLowToneFilt   dsm SymbolLen
RxHighToneFilt  dsm SymbolLen
RxTonePowerPipe dsm 2*SymbolLen
RxSymbolSyncTap dsm SymbolLen


;*************************************************************************
;External L (X/Y) data RAM

        if EVM56K
          LOMEM X:$2000,Y:$2000,L:$2000
          HIMEM X:$3FFF,Y:$3FFF,L:$3FFF
        else
          LOMEM X:$0200,Y:$0200,L:$0200
          HIMEM X:$1FFF,Y:$1FFF,L:$1FFF
        endif

      if EVM56K
        org L:$2000
      else
        org L:$200
      endif

        org L:          ;external L-RAM

LastL = *
        org X:LastL
        org Y:LastL

        org Y:          ;external Y-RAM

AliasFilterInpI dsm AliasFilterLen      ;anti-alias input filter (I/Q)
AliasFilterInpQ dsm AliasFilterLen

        org X:          ;external X-RAM

AliasFilterOutI dsm AliasFilterLen      ;anti-alias output filter (I/Q)
AliasFilterOutQ dsm AliasFilterLen

        org X:
LastX = *
        org Y:
LastY = *
        if @cvi(LastX)>=@cvi(LastY)
          org L:LastX
        else
          org L:LastY
        endif

Buffer  dsm BufLen*4    ;CODEC's input/output buffer

;*************************************************************************
                        ;external Y-only  RAM

        if EVM56K
          LOMEM Y:$0200
          HIMEM Y:$1FFF
        else
          LOMEM Y:$2000
          HIMEM Y:$3FFF
        endif

      if EVM56K
        org Y:$200
      else
        org Y:$2000
      endif

        org Y:

;*************************************************************************
;constant tables: FIR and window shapes, FFT coefficiants, etc.

        if EVM56K
          LOMEM X:$2000,Y:$0200,L:$2000
          HIMEM X:$3FFF,Y:$3FFF,L:$3FFF
        else
          LOMEM X:$0200,Y:$0200,L:$0200
          HIMEM X:$1FFF,Y:$3FFF,L:$1FFF
        endif



LowFreq equ AliasLowFreq/SampleFreq
UppFreq equ AliasUppFreq/SampleFreq
CenterFreq equ (LowFreq+UppFreq)/2
SideFreq  equ (UppFreq-LowFreq)/2

w0      equ     0.35875 ;coeff. for Blackman-Harris 4 term
w1      equ     0.48829 ;minimum sidelobe window (copied from Motorola BBS)
w2      equ     0.14128
w3	equ	0.01168

        org Y:AliasFilterInpI

time = -@cvf(AliasFilterLen/2)+0.5
count   set 0
        dup AliasFilterLen
angle = PI*time/@cvf(AliasFilterLen/2)
Window = w0+w1*@cos(angle)+w2*@cos(2.0*angle)+w3*@cos(3.0*angle)
Filter = @sin(2.0*PI*time*SideFreq)/(PI*time)
        dc @cos(2.0*PI*time*CenterFreq)*Window*Filter
time = time+1.0
count   set count+1
        endm

        org Y:AliasFilterInpQ

time = -@cvf(AliasFilterLen/2)+0.5
count   set 0
        dup AliasFilterLen
angle = PI*time/@cvf(AliasFilterLen/2)
Window = w0+w1*@cos(angle)+w2*@cos(2.0*angle)+w3*@cos(3.0*angle)
Filter = @sin(2.0*PI*time*SideFreq)/(PI*time)
        dc -@sin(2.0*PI*time*CenterFreq)*Window*Filter
time = time+1.0
count   set count+1
        endm

        org X:AliasFilterOutI

time = -@cvf(AliasFilterLen/2)+0.5
count   set 0
        dup AliasFilterLen
angle = PI*time/@cvf(AliasFilterLen/2)
Window = w0+w1*@cos(angle)+w2*@cos(2.0*angle)+w3*@cos(3.0*angle)
Filter = @sin(2.0*PI*time*SideFreq)/(PI*time)
        dc @cos(2.0*PI*time*CenterFreq)*Window*Filter
time = time+1.0
count   set count+1
        endm

        org X:AliasFilterOutQ

time = -@cvf(AliasFilterLen/2)+0.5
count   set 0
        dup AliasFilterLen
angle = PI*time/@cvf(AliasFilterLen/2)
Window = w0+w1*@cos(angle)+w2*@cos(2.0*angle)+w3*@cos(3.0*angle)
Filter = @sin(2.0*PI*time*SideFreq)/(PI*time)
        dc -@sin(2.0*PI*time*CenterFreq)*Window*Filter
time = time+1.0
count   set count+1
        endm

;*************************************************************************

        end

